from typing import Any, Literal

from kirin import rewrite
from kirin.prelude import basic
from kirin.dialects import py, ilist


def test():
    rule = rewrite.Fixpoint(
        rewrite.Walk(rewrite.Chain(ilist.rewrite.HintLen(), rewrite.ConstantFold()))
    )

    @basic
    def len_func(xs: ilist.IList[int, Literal[3]]):
        return len(xs)

    @basic
    def len_func3(xs: ilist.IList[int, Any]):
        return len(xs)

    rule.rewrite(len_func.code)
    rule.rewrite(len_func3.code)

    stmt = len_func.callable_region.blocks[0].stmts.at(0)
    assert isinstance(stmt, py.Constant)
    assert stmt.value.unwrap() == 3

    stmt = len_func3.callable_region.blocks[0].stmts.at(0)
    assert isinstance(stmt, py.Len)



from kirin.prelude import basic


@basic
def foo(x: int):  # type: ignore
    def goo(y: int):
        return x + y

    return goo


@basic
def main(y: int):
    x = 1
    f = foo(x)
    return f(y)


def test_main():
    assert main(1) == 2
    assert main(2) == 3



from typing import TYPE_CHECKING

from kirin import ir

if TYPE_CHECKING:
    from kirin.ir import Statement
    from kirin.print.printer import Printer


def pprint_calllike(
    invoke_or_call: "Statement", callee: str, printer: "Printer"
) -> None:
    with printer.rich(style="red"):
        printer.print_name(invoke_or_call)
    printer.plain_print(" ")

    n_total = len(invoke_or_call.args)
    printer.plain_print(callee)
    if (inputs := getattr(invoke_or_call, "inputs", None)) is None:
        raise ValueError(f"{invoke_or_call} does not have inputs")

    if not isinstance(inputs, tuple):
        raise ValueError(f"inputs of {invoke_or_call} is not a tuple")

    if (kwargs := getattr(invoke_or_call, "kwargs", None)) is None:
        raise ValueError(f"{invoke_or_call} does not have kwargs")

    if not isinstance(kwargs, tuple):
        raise ValueError(f"kwargs of {invoke_or_call} is not a tuple")

    positional = inputs[: n_total - len(kwargs)]
    kwargs = dict(
        zip(
            kwargs,
            inputs[n_total - len(kwargs) :],
        )
    )

    printer.plain_print("(")
    printer.print_seq(positional)
    if kwargs and positional:
        printer.plain_print(", ")
    printer.print_mapping(kwargs, delim=", ")
    printer.plain_print(")")

    with printer.rich(style="comment"):
        printer.plain_print(" : ")
        printer.print_seq(
            [result.type for result in invoke_or_call._results],
            delim=", ",
        )
        if trait := invoke_or_call.get_trait(ir.MaybePure):
            printer.plain_print(f" maybe_pure={trait.is_pure(invoke_or_call)}")



from dataclasses import dataclass

from kirin.ir import Method
from kirin.rewrite import Walk, WrapConst
from kirin.analysis import const
from kirin.passes.abc import Pass
from kirin.rewrite.abc import RewriteResult


@dataclass
class HintConst(Pass):

    def unsafe_run(self, mt: Method) -> RewriteResult:
        constprop = const.Propagate(self.dialects)
        frame, _ = constprop.run_analysis(mt, no_raise=self.no_raise)
        return Walk(WrapConst(frame)).rewrite(mt.code)



from kirin import types
from kirin.prelude import basic


@basic
def tuple_hint(xs: tuple[int, ...]):
    types.Tuple[types.Int]


def test_tuple_hint():
    assert tuple_hint.arg_types[0].is_subseteq(types.Tuple[types.Vararg(types.Int)])



from kirin.prelude import basic


@basic
def foo(x: float):
    return x + 0.22


@basic
def issue_87(x: float):

    def inner(y: float, z: float):
        return foo(x) + y + z

    return inner


def test_issue_87():
    assert issue_87(1.0)(1, 2) == 4.22



import pytest

from kirin import ir
from kirin.decl import info, statement

dialect = ir.Dialect("my_dialect")


def test_reserved_verify():
    with pytest.raises(ValueError):

        @statement(dialect=dialect)
        class ReserveKeyword(ir.Statement):
            name = "my_statement"
            traits = frozenset({})
            args: ir.SSAValue = info.argument()

    with pytest.raises(ValueError):

        @statement(dialect=dialect)
        class NoAnnotation(ir.Statement):
            name = "my_statement"
            traits = frozenset({})
            no_annotation = info.argument()  # type: ignore

    with pytest.raises(ValueError):

        @statement(dialect=dialect)
        class WrongAnnotation(ir.Statement):
            name = "my_statement"
            traits = frozenset({})
            field: str = info.argument()

    with pytest.raises(ValueError):

        @statement(dialect=dialect)
        class WrongFiledSpecifier(ir.Statement):
            name = "my_statement"
            traits = frozenset({})
            result: ir.ResultValue = info.argument()

    with pytest.raises(ValueError):

        @statement(dialect=dialect)
        class WrongResultAnnotation(ir.Statement):
            name = "my_statement"
            traits = frozenset({})
            result: ir.SSAValue = info.result()



from kirin.prelude import basic_no_opt
from kirin.analysis import const


def test_worklist_bfs():
    @basic_no_opt
    def make_ker(val: float):

        def ker(i: float):
            return i + val

        return ker

    @basic_no_opt
    def test(x: str, y: float, flag: bool):

        if x == "x":
            val = 1.0
        else:
            val = 2.0

        if flag:
            ker = make_ker(val=val)

        else:
            ker = make_ker(val=val)

        return ker

    # test.print()
    prop = const.Propagate(basic_no_opt)
    frame, ret = prop.run_analysis(test)
    assert isinstance(ret, const.PartialLambda)



from typing import Generic, TypeVar
from dataclasses import field, dataclass

from kirin.interp.frame import FrameABC

FrameType = TypeVar("FrameType", bound=FrameABC)


@dataclass
class InterpreterState(Generic[FrameType]):
    """Interpreter state.

    This class represents the state of the interpreter. It contains the
    stack of frames for the interpreter. The stack of frames is used to
    store the current state of the interpreter during interpretation.
    """

    _current_frame: FrameType | None = field(
        default=None, kw_only=True, init=False, repr=False
    )
    """Current frame of the stack."""
    depth: int = field(default=0, kw_only=True, init=False, repr=False)
    """stack depth of the interpreter."""

    @property
    def current_frame(self) -> FrameType:
        """Get the current frame.

        Returns:
            FrameType: The current frame.
        """
        if self._current_frame is None:
            raise ValueError("no current frame")
        return self._current_frame

    def push_frame(self, frame: FrameType) -> FrameType:
        """Push a frame onto the stack.

        Args:
            frame(FrameType): The frame to push onto the stack.

        Returns:
            FrameType: The frame that was pushed.
        """
        assert frame.parent is None, "frame already has a parent"
        self.depth += 1
        if self._current_frame is None:
            self._current_frame = frame
        else:
            frame.parent = self._current_frame
            self._current_frame = frame
        return self._current_frame

    def pop_frame(self) -> FrameType:
        """Pop a frame from the stack.

        Returns:
            FrameType: The frame that was popped.
        """
        if self._current_frame is None:
            raise ValueError("no frame to pop")
        frame = self._current_frame
        self._current_frame = self._current_frame.parent
        self.depth -= 1
        frame.parent = None
        return frame



from __future__ import annotations

import ast
import sys
from abc import ABC
from typing import Any, Callable, ClassVar, TypeAlias
from dataclasses import dataclass

from kirin.ir import Method, SSAValue
from kirin.ir.attrs import types
from kirin.lowering.abc import Result
from kirin.lowering.state import State

LoweringTransform: TypeAlias = Callable[[Any, State[ast.AST], ast.Call], Result]

@dataclass
class Transform:
    objs: tuple[Callable, ...]
    func: LoweringTransform

@dataclass
class akin:
    obj: Callable

    def __call__(
        self,
        func: LoweringTransform | Transform,
    ) -> Transform: ...

class FromPythonAST(ABC):
    callee_table: ClassVar[dict[object, Transform]]

    @property
    def names(self) -> list[str]: ...
    def lower(self, state: State[ast.AST], node: ast.AST) -> Result: ...
    def unreachable(self, state: State[ast.AST], node: ast.AST) -> Result: ...
    def get_hint(
        self, state: State[ast.AST], node: ast.expr | None
    ) -> types.TypeAttribute: ...
    def lower_Module(self, state: State[ast.AST], node: ast.Module) -> Result: ...
    def lower_Interactive(
        self, state: State[ast.AST], node: ast.Interactive
    ) -> Result: ...
    def lower_Expression(
        self, state: State[ast.AST], node: ast.Expression
    ) -> Result: ...
    def lower_FunctionDef(
        self, state: State[ast.AST], node: ast.FunctionDef
    ) -> Result: ...
    def lower_AsyncFunctionDef(
        self, state: State[ast.AST], node: ast.AsyncFunctionDef
    ) -> Result: ...
    def lower_ClassDef(self, state: State[ast.AST], node: ast.ClassDef) -> Result: ...
    def lower_Return(self, state: State[ast.AST], node: ast.Return) -> Result: ...
    def lower_Delete(self, state: State[ast.AST], node: ast.Delete) -> Result: ...
    def lower_Assign(self, state: State[ast.AST], node: ast.Assign) -> Result: ...
    def lower_AugAssign(self, state: State[ast.AST], node: ast.AugAssign) -> Result: ...
    def lower_AnnAssign(self, state: State[ast.AST], node: ast.AnnAssign) -> Result: ...
    def lower_For(self, state: State[ast.AST], node: ast.For) -> Result: ...
    def lower_AsyncFor(self, state: State[ast.AST], node: ast.AsyncFor) -> Result: ...
    def lower_While(self, state: State[ast.AST], node: ast.While) -> Result: ...
    def lower_If(self, state: State[ast.AST], node: ast.If) -> Result: ...
    def lower_With(self, state: State[ast.AST], node: ast.With) -> Result: ...
    def lower_AsyncWith(self, state: State[ast.AST], node: ast.AsyncWith) -> Result: ...
    def lower_Raise(self, state: State[ast.AST], node: ast.Raise) -> Result: ...
    def lower_Try(self, state: State[ast.AST], node: ast.Try) -> Result: ...
    def lower_Assert(self, state: State[ast.AST], node: ast.Assert) -> Result: ...
    def lower_Import(self, state: State[ast.AST], node: ast.Import) -> Result: ...
    def lower_ImportFrom(
        self, state: State[ast.AST], node: ast.ImportFrom
    ) -> Result: ...
    def lower_Global(self, state: State[ast.AST], node: ast.Global) -> Result: ...
    def lower_Nonlocal(self, state: State[ast.AST], node: ast.Nonlocal) -> Result: ...
    def lower_Expr(self, state: State[ast.AST], node: ast.Expr) -> Result: ...
    def lower_Pass(self, state: State[ast.AST], node: ast.Pass) -> Result: ...
    def lower_Break(self, state: State[ast.AST], node: ast.Break) -> Result: ...
    def lower_Continue(self, state: State[ast.AST], node: ast.Continue) -> Result: ...
    def lower_Slice(self, state: State[ast.AST], node: ast.Slice) -> Result: ...
    def lower_BoolOp(self, state: State[ast.AST], node: ast.BoolOp) -> Result: ...
    def lower_BinOp(self, state: State[ast.AST], node: ast.BinOp) -> Result: ...
    def lower_UnaryOp(self, state: State[ast.AST], node: ast.UnaryOp) -> Result: ...
    def lower_Lambda(self, state: State[ast.AST], node: ast.Lambda) -> Result: ...
    def lower_IfExp(self, state: State[ast.AST], node: ast.IfExp) -> Result: ...
    def lower_Dict(self, state: State[ast.AST], node: ast.Dict) -> Result: ...
    def lower_Set(self, state: State[ast.AST], node: ast.Set) -> Result: ...
    def lower_ListComp(self, state: State[ast.AST], node: ast.ListComp) -> Result: ...
    def lower_SetComp(self, state: State[ast.AST], node: ast.SetComp) -> Result: ...
    def lower_DictComp(self, state: State[ast.AST], node: ast.DictComp) -> Result: ...
    def lower_GeneratorExp(
        self, state: State[ast.AST], node: ast.GeneratorExp
    ) -> Result: ...
    def lower_Await(self, state: State[ast.AST], node: ast.Await) -> Result: ...
    def lower_Yield(self, state: State[ast.AST], node: ast.Yield) -> Result: ...
    def lower_YieldFrom(self, state: State[ast.AST], node: ast.YieldFrom) -> Result: ...
    def lower_Compare(self, state: State[ast.AST], node: ast.Compare) -> Result: ...
    def lower_Call(self, state: State[ast.AST], node: ast.Call) -> Result: ...
    def lower_Call_global_method(
        self, state: State[ast.AST], method: Method, node: ast.Call
    ) -> Result: ...
    def lower_Call_local(
        self, state: State[ast.AST], callee: SSAValue, node: ast.Call
    ) -> Result: ...
    def lower_FormattedValue(
        self, state: State[ast.AST], node: ast.FormattedValue
    ) -> Result: ...
    def lower_JoinedStr(self, state: State[ast.AST], node: ast.JoinedStr) -> Result: ...
    def lower_Constant(self, state: State[ast.AST], node: ast.Constant) -> Result: ...
    def lower_NamedExpr(self, state: State[ast.AST], node: ast.NamedExpr) -> Result: ...
    def lower_TypeIgnore(
        self, state: State[ast.AST], node: ast.TypeIgnore
    ) -> Result: ...
    def lower_Attribute(self, state: State[ast.AST], node: ast.Attribute) -> Result: ...
    def lower_Subscript(self, state: State[ast.AST], node: ast.Subscript) -> Result: ...
    def lower_Starred(self, state: State[ast.AST], node: ast.Starred) -> Result: ...
    def lower_Name(self, state: State[ast.AST], node: ast.Name) -> Result: ...
    def lower_List(self, state: State[ast.AST], node: ast.List) -> Result: ...
    def lower_Tuple(self, state: State[ast.AST], node: ast.Tuple) -> Result: ...
    def lower_Del(self, state: State[ast.AST], node: ast.Del) -> Result: ...
    def lower_Load(self, state: State[ast.AST], node: ast.Load) -> Result: ...
    def lower_Store(self, state: State[ast.AST], node: ast.Store) -> Result: ...
    def lower_And(self, state: State[ast.AST], node: ast.And) -> Result: ...
    def lower_Or(self, state: State[ast.AST], node: ast.Or) -> Result: ...
    def lower_Add(self, state: State[ast.AST], node: ast.Add) -> Result: ...
    def lower_BitAnd(self, state: State[ast.AST], node: ast.BitAnd) -> Result: ...
    def lower_BitOr(self, state: State[ast.AST], node: ast.BitOr) -> Result: ...
    def lower_BitXor(self, state: State[ast.AST], node: ast.BitXor) -> Result: ...
    def lower_Div(self, state: State[ast.AST], node: ast.Div) -> Result: ...
    def lower_FloorDiv(self, state: State[ast.AST], node: ast.FloorDiv) -> Result: ...
    def lower_LShift(self, state: State[ast.AST], node: ast.LShift) -> Result: ...
    def lower_Mod(self, state: State[ast.AST], node: ast.Mod) -> Result: ...
    def lower_Mult(self, state: State[ast.AST], node: ast.Mult) -> Result: ...
    def lower_MatMult(self, state: State[ast.AST], node: ast.MatMult) -> Result: ...
    def lower_Pow(self, state: State[ast.AST], node: ast.Pow) -> Result: ...
    def lower_RShift(self, state: State[ast.AST], node: ast.RShift) -> Result: ...
    def lower_Sub(self, state: State[ast.AST], node: ast.Sub) -> Result: ...
    def lower_Invert(self, state: State[ast.AST], node: ast.Invert) -> Result: ...
    def lower_Not(self, state: State[ast.AST], node: ast.Not) -> Result: ...
    def lower_UAdd(self, state: State[ast.AST], node: ast.UAdd) -> Result: ...
    def lower_USub(self, state: State[ast.AST], node: ast.USub) -> Result: ...
    def lower_Eq(self, state: State[ast.AST], node: ast.Eq) -> Result: ...
    def lower_Gt(self, state: State[ast.AST], node: ast.Gt) -> Result: ...
    def lower_GtE(self, state: State[ast.AST], node: ast.GtE) -> Result: ...
    def lower_In(self, state: State[ast.AST], node: ast.In) -> Result: ...
    def lower_Is(self, state: State[ast.AST], node: ast.Is) -> Result: ...
    def lower_IsNot(self, state: State[ast.AST], node: ast.IsNot) -> Result: ...
    def lower_Lt(self, state: State[ast.AST], node: ast.Lt) -> Result: ...
    def lower_LtE(self, state: State[ast.AST], node: ast.LtE) -> Result: ...
    def lower_NotEq(self, state: State[ast.AST], node: ast.NotEq) -> Result: ...
    def lower_NotIn(self, state: State[ast.AST], node: ast.NotIn) -> Result: ...
    def lower_comprehension(
        self, state: State[ast.AST], node: ast.comprehension
    ) -> Result: ...
    def lower_ExceptHandler(
        self, state: State[ast.AST], node: ast.ExceptHandler
    ) -> Result: ...
    def lower_arguments(self, state: State[ast.AST], node: ast.arguments) -> Result: ...
    def lower_arg(self, state: State[ast.AST], node: ast.arg) -> Result: ...
    def lower_keyword(self, state: State[ast.AST], node: ast.keyword) -> Result: ...
    def lower_alias(self, state: State[ast.AST], node: ast.alias) -> Result: ...
    def lower_withitem(self, state: State[ast.AST], node: ast.withitem) -> Result: ...
    if sys.version_info >= (3, 10):
        def lower_Match(self, state: State[ast.AST], node: ast.Match) -> Result: ...
        def lower_match_case(
            self, state: State[ast.AST], node: ast.match_case
        ) -> Result: ...
        def lower_MatchValue(
            self, state: State[ast.AST], node: ast.MatchValue
        ) -> Result: ...
        def lower_MatchSequence(
            self, state: State[ast.AST], node: ast.MatchSequence
        ) -> Result: ...
        def lower_MatchSingleton(
            self, state: State[ast.AST], node: ast.MatchSingleton
        ) -> Result: ...
        def lower_MatchStar(
            self, state: State[ast.AST], node: ast.MatchStar
        ) -> Result: ...
        def lower_MatchMapping(
            self, state: State[ast.AST], node: ast.MatchMapping
        ) -> Result: ...
        def lower_MatchClass(
            self, state: State[ast.AST], node: ast.MatchClass
        ) -> Result: ...
        def lower_MatchAs(self, state: State[ast.AST], node: ast.MatchAs) -> Result: ...
        def lower_MatchOr(self, state: State[ast.AST], node: ast.MatchOr) -> Result: ...

    if sys.version_info >= (3, 11):
        def lower_TryStar(self, state: State[ast.AST], node: ast.TryStar) -> Result: ...

    if sys.version_info >= (3, 12):
        def lower_TypeVar(self, state: State[ast.AST], node: ast.TypeVar) -> Result: ...
        def lower_ParamSpec(
            self, state: State[ast.AST], node: ast.ParamSpec
        ) -> Result: ...
        def lower_TypeVarTuple(
            self, state: State[ast.AST], node: ast.TypeVarTuple
        ) -> Result: ...
        def lower_TypeAlias(
            self, state: State[ast.AST], node: ast.TypeAlias
        ) -> Result: ...

class NoSpecialLowering(FromPythonAST):
    pass



# re-exports the public API of the kirin package
from . import ir as ir, types as types, lowering as lowering
from .exception import enable_stracetrace, disable_stracetrace

__all__ = ["ir", "types", "lowering", "enable_stracetrace", "disable_stracetrace"]



import ast

from kirin import lowering

from . import stmts
from ._dialect import dialect


@dialect.register
class Lowering(lowering.FromPythonAST):

    def lower_UnaryOp(
        self, state: lowering.State, node: ast.UnaryOp
    ) -> lowering.Result:
        if op := getattr(stmts, node.op.__class__.__name__, None):
            return state.current_frame.push(op(state.lower(node.operand).expect_one()))
        else:
            raise lowering.BuildError(f"unsupported unary operator {node.op}")



from dataclasses import field, dataclass

from attrs import Food
from stmts import Nap, Cook, NewFood
from dialect import dialect

from kirin import ir, interp
from lattice import Item, ItemFood, ItemServing, AtLeastXItem, ConstIntItem
from kirin.interp import exceptions
from kirin.analysis import Forward
from kirin.dialects import py
from kirin.dialects.py import binop
from kirin.analysis.forward import ForwardFrame


@dataclass
class FeeAnalysis(Forward[Item]):
    keys = ["food.fee"]
    lattice = Item
    nap_count: int = field(init=False)

    def initialize(self):
        """Initialize the analysis pass.

        The method is called before the analysis pass starts.

        Note:
            1. Here one is *required* to call the super().initialize() to initialize the analysis pass,
            which clear all the previous analysis results and symbol tables.
            2. Any additional initialization that belongs to the analysis should also be done here.
            For example, in this case, we initialize the nap_count to 0.

        """
        super().initialize()
        self.nap_count = 0
        return self

    def eval_stmt_fallback(
        self, frame: ForwardFrame[Item], stmt: ir.Statement
    ) -> tuple[Item, ...] | interp.SpecialValue[Item]:
        return ()

    def run_method(self, method: ir.Method, args: tuple[Item, ...]):
        return self.run_callable(method.code, (self.lattice.bottom(),) + args)


@py.constant.dialect.register(key="food.fee")
class PyConstMethodTable(interp.MethodTable):

    @interp.impl(py.constant.Constant)
    def const(
        self,
        interp: FeeAnalysis,
        frame: interp.Frame[Item],
        stmt: py.constant.Constant,
    ):
        if isinstance(stmt.value, int):
            return (ConstIntItem(data=stmt.value),)
        elif isinstance(stmt.value, Food):
            return (ItemFood(type=stmt.value.type),)

        else:
            raise exceptions.InterpreterError(
                f"illegal constant type {type(stmt.value)}"
            )


@binop.dialect.register(key="food.fee")
class PyBinOpMethodTable(interp.MethodTable):

    @interp.impl(binop.Add)
    def add(
        self,
        interp: FeeAnalysis,
        frame: interp.Frame[Item],
        stmt: binop.Add,
    ):
        left = frame.get(stmt.lhs)
        right = frame.get(stmt.rhs)

        if isinstance(left, AtLeastXItem) or isinstance(right, AtLeastXItem):
            out = AtLeastXItem(data=left.data + right.data)
        else:
            out = ConstIntItem(data=left.data + right.data)

        return (out,)


@dialect.register(key="food.fee")
class FoodMethodTable(interp.MethodTable):

    @interp.impl(NewFood)
    def new_food(
        self,
        interp: FeeAnalysis,
        frame: interp.Frame[Item],
        stmt: NewFood,
    ):
        return (ItemFood(type=stmt.type),)

    @interp.impl(Cook)
    def cook(
        self,
        interp: FeeAnalysis,
        frame: interp.Frame[Item],
        stmt: Cook,
    ):
        # food depends on the food type to have different charge:

        food = frame.get_typed(stmt.target, ItemFood)
        serving_count: AtLeastXItem | ConstIntItem = frame.get(
            stmt.amount
        )  # FIXME: fix the type hinting here

        out = ItemServing(count=serving_count, type=food.type)

        return (out,)

    @interp.impl(Nap)
    def nap(
        self,
        interp: FeeAnalysis,
        frame: interp.Frame[Item],
        stmt: Nap,
    ):
        interp.nap_count += 1
        return ()



import pytest

from kirin import ir, lowering
from kirin.decl import info, statement
from kirin.prelude import basic_no_opt
from kirin.dialects import cf, py

dialect = ir.Dialect("test")


@statement(dialect=dialect)
class MultiResult(ir.Statement):
    traits = frozenset({lowering.FromPythonCall()})
    result_a: ir.ResultValue = info.result()
    result_b: ir.ResultValue = info.result()


dummy_dialect = basic_no_opt.add(dialect)


def test_multi_result():
    @dummy_dialect
    def multi_assign():
        (x, y) = MultiResult()  # type: ignore
        return x, y

    stmt = multi_assign.callable_region.blocks[0].stmts.at(0)
    assert isinstance(stmt, MultiResult)
    assert stmt.result_a.name == "x"
    assert stmt.result_b.name == "y"

    with pytest.raises(lowering.BuildError):

        @dummy_dialect
        def multi_assign_error():
            (x, y, z) = MultiResult()  # type: ignore
            return x, y, z


def test_chain_assign_setattr():

    @dummy_dialect
    def chain_assign(y):
        x = y.z = 1
        return x, y

    stmt = chain_assign.callable_region.blocks[0].stmts.at(1)
    assert isinstance(stmt, py.assign.SetAttribute)
    assert stmt.obj.name == "y"
    assert stmt.attr == "z"
    assert stmt.value.name == "x"


def test_aug_assign():
    @dummy_dialect
    def aug_assign(y):
        y += 1
        return y

    y = aug_assign.callable_region.blocks[0].args[1]
    const = aug_assign.callable_region.blocks[0].stmts.at(0)
    assert isinstance(const, py.Constant)
    assert const.value.unwrap() == 1
    add = aug_assign.callable_region.blocks[0].stmts.at(1)
    assert isinstance(add, py.binop.Add)
    assert add.lhs is y
    assert add.rhs is const.result


def test_named_expr():

    @dummy_dialect
    def named_expr(y):
        if y := y + 1:
            return y
        return y

    stmt = named_expr.callable_region.blocks[0].stmts.at(1)
    y = named_expr.callable_region.blocks[0].args[1]
    assert isinstance(stmt, py.binop.Add)
    assert stmt.lhs is y
    br = named_expr.callable_region.blocks[0].stmts.at(2)
    assert isinstance(br, cf.ConditionalBranch)
    assert stmt.result is br.cond



from kirin import interp

from . import stmts as cmp
from ._dialect import dialect


@dialect.register
class CmpMethod(interp.MethodTable):

    @interp.impl(cmp.Eq)
    def eq(self, interp, frame: interp.Frame, stmt: cmp.Eq):
        return (frame.get(stmt.lhs) == frame.get(stmt.rhs),)

    @interp.impl(cmp.NotEq)
    def not_eq(self, interp, frame: interp.Frame, stmt: cmp.NotEq):
        return (frame.get(stmt.lhs) != frame.get(stmt.rhs),)

    @interp.impl(cmp.Lt)
    def lt(self, interp, frame: interp.Frame, stmt: cmp.Lt):
        return (frame.get(stmt.lhs) < frame.get(stmt.rhs),)

    @interp.impl(cmp.LtE)
    def lt_eq(self, interp, frame: interp.Frame, stmt: cmp.LtE):
        return (frame.get(stmt.lhs) <= frame.get(stmt.rhs),)

    @interp.impl(cmp.Gt)
    def gt(self, interp, frame: interp.Frame, stmt: cmp.Gt):
        return (frame.get(stmt.lhs) > frame.get(stmt.rhs),)

    @interp.impl(cmp.GtE)
    def gt_eq(self, interp, frame: interp.Frame, stmt: cmp.GtE):
        return (frame.get(stmt.lhs) >= frame.get(stmt.rhs),)

    @interp.impl(cmp.In)
    def in_(self, interp, frame: interp.Frame, stmt: cmp.In):
        return (frame.get(stmt.lhs) in frame.get(stmt.rhs),)

    @interp.impl(cmp.NotIn)
    def not_in(self, interp, frame: interp.Frame, stmt: cmp.NotIn):
        return (frame.get(stmt.lhs) not in frame.get(stmt.rhs),)

    @interp.impl(cmp.Is)
    def is_(self, interp, frame: interp.Frame, stmt: cmp.Is):
        return (frame.get(stmt.lhs) is frame.get(stmt.rhs),)

    @interp.impl(cmp.IsNot)
    def is_not(self, interp, frame: interp.Frame, stmt: cmp.IsNot):
        return (frame.get(stmt.lhs) is not frame.get(stmt.rhs),)



from typing import Any

from kirin.ir import Region
from kirin.ir.method import Method
from kirin.ir.nodes.stmt import Statement

from .base import BaseInterpreter
from .frame import Frame
from .value import Successor, YieldValue, ReturnValue, SpecialValue
from .exceptions import FuelExhaustedError


class Interpreter(BaseInterpreter[Frame[Any], Any]):
    """Concrete interpreter for the IR.

    This is a concrete interpreter for the IR. It evaluates the IR by
    executing the statements in the IR using a simple stack-based
    interpreter.
    """

    keys = ["main"]
    void = None

    def initialize_frame(
        self, code: Statement, *, has_parent_access: bool = False
    ) -> Frame[Any]:
        return Frame(code, has_parent_access=has_parent_access)

    def run_method(
        self, method: Method, args: tuple[Any, ...]
    ) -> tuple[Frame[Any], Any]:
        return self.run_callable(method.code, (method,) + args)

    def run_ssacfg_region(
        self, frame: Frame[Any], region: Region, args: tuple[Any, ...]
    ) -> tuple[Any, ...] | None | ReturnValue[Any]:
        block = region.blocks[0]
        succ = Successor(block, *args)
        while succ is not None:
            results = self.run_succ(frame, succ)
            if isinstance(results, Successor):
                succ = results
            elif isinstance(results, ReturnValue):
                return results
            elif isinstance(results, YieldValue):
                return results.values
            else:
                return results
        return None  # region without terminator returns empty tuple

    def run_succ(self, frame: Frame[Any], succ: Successor) -> SpecialValue[Any]:
        frame.current_block = succ.block
        frame.set_values(succ.block.args, succ.block_args)
        for stmt in succ.block.stmts:
            if self.consume_fuel() == self.FuelResult.Stop:
                raise FuelExhaustedError("fuel exhausted")
            frame.current_stmt = stmt
            stmt_results = self.eval_stmt(frame, stmt)
            if isinstance(stmt_results, tuple):
                frame.set_values(stmt._results, stmt_results)
            elif stmt_results is None:
                continue  # empty result
            else:  # terminator
                return stmt_results
        return None



from abc import ABC, ABCMeta, abstractmethod
from typing import Generic, TypeVar, Iterable


class LatticeMeta(ABCMeta):
    pass


class SingletonMeta(LatticeMeta):
    """
    Singleton metaclass for lattices. It ensures that only one instance of a lattice is created.

    See https://stackoverflow.com/questions/674304/why-is-init-always-called-after-new/8665179#8665179
    """

    def __init__(cls, name, bases, attrs):
        super().__init__(name, bases, attrs)
        cls._instance = None

    def __call__(cls):
        if cls._instance is None:
            cls._instance = super().__call__()
        return cls._instance


LatticeType = TypeVar("LatticeType", bound="Lattice")


class Lattice(ABC, Generic[LatticeType], metaclass=LatticeMeta):
    """ABC for lattices as Python class.

    While `Lattice` is only an interface, `LatticeABC` is an abstract
    class that can be inherited from. This provides a few default
    implementations for the lattice operations.
    """

    @abstractmethod
    def join(self, other: LatticeType) -> LatticeType:
        """Join operation."""
        ...

    @abstractmethod
    def meet(self, other: LatticeType) -> LatticeType:
        """Meet operation."""
        ...

    @abstractmethod
    def is_subseteq(self, other: LatticeType) -> bool:
        """Subseteq operation."""
        ...

    def is_equal(self, other: LatticeType) -> bool:
        """Check if two lattices are equal."""
        if self is other:
            return True
        else:
            return self.is_subseteq(other) and other.is_subseteq(self)

    def is_subset(self, other: LatticeType) -> bool:
        return self.is_subseteq(other) and not other.is_subseteq(self)

    def __eq__(self, value: object) -> bool:
        raise NotImplementedError(
            "Equality is not implemented for lattices, use is_equal instead"
        )

    def __hash__(self) -> int:
        raise NotImplementedError("Hash is not implemented for lattices")


BoundedLatticeType = TypeVar("BoundedLatticeType", bound="BoundedLattice")


class BoundedLattice(Lattice[BoundedLatticeType]):
    """ABC for bounded lattices as Python class.

    `BoundedLattice` is an abstract class that can be inherited from.
    It requires the implementation of the `bottom` and `top` methods.
    """

    @classmethod
    @abstractmethod
    def bottom(cls) -> BoundedLatticeType: ...

    @classmethod
    @abstractmethod
    def top(cls) -> BoundedLatticeType: ...


class UnionMeta(LatticeMeta):
    """Meta class for union types. It simplifies the union if possible."""

    def __init__(self, name, bases, attrs):
        super().__init__(name, bases, attrs)
        if not issubclass(base := bases[0], BoundedLattice):
            raise TypeError(f"Union must inherit from Lattice, got {bases[0]}")
        self._bottom = base.bottom()

    def __call__(
        self,
        typ: Iterable[LatticeType] | LatticeType,
        *others: LatticeType,
    ):
        from kirin.lattice.abc import Lattice

        if isinstance(typ, Lattice):
            typs: Iterable[LatticeType] = (typ, *others)
        elif not others:
            typs = typ
        else:
            raise ValueError(
                "Expected an iterable of types or variadic arguments of types"
            )

        # try if the union can be simplified
        params: list[LatticeType] = []
        for typ in typs:
            contains = False
            for idx, other in enumerate(params):
                if typ.is_subseteq(other):
                    contains = True
                    break
                elif other.is_subseteq(typ):
                    params[idx] = typ
                    contains = True
                    break

            if not contains:
                params.append(typ)

        if len(params) == 1:
            return params[0]

        if len(params) == 0:
            return self._bottom
        return super(UnionMeta, self).__call__(*params)



from typing import Any, Generator
from contextlib import contextmanager

from kirin import ir, lowering
from kirin.decl import info, statement
from kirin.prelude import structural_no_opt
from kirin.dialects import ilist

dialect = ir.Dialect("with_binding")


@statement(dialect=dialect)
class ContextStatatement(ir.Statement):
    traits = frozenset({lowering.FromPythonWithSingleItem()})
    body: ir.Region = info.region(multi=False)


@ir.dialect_group(structural_no_opt.add(dialect))
def dummy(self):

    def run_pass(mt):

        return mt

    return run_pass


@lowering.wraps(ContextStatatement)
@contextmanager
def context_statement() -> Generator[Any, None, None]: ...


@dummy
def with_binding():
    x = 1

    def fn(x):
        return x**2

    with context_statement():
        with context_statement():
            x = ilist.map(fn, ilist.range(10))

    return x


def test_with_binding():
    stmt = with_binding.callable_region.blocks[0].stmts.at(-2)
    assert isinstance(stmt, ContextStatatement)
    assert len(stmt.body.blocks) == 1
    stmt = stmt.body.blocks[0].stmts.at(0)
    assert isinstance(stmt, ContextStatatement)
    assert len(stmt.body.blocks[0].stmts) == 5



from kirin import interp

from . import stmts
from ._dialect import dialect


@dialect.register
class PyMethodTable(interp.MethodTable):

    @interp.impl(stmts.Add)
    def add(self, interp, frame: interp.Frame, stmt: stmts.Add):
        return (frame.get(stmt.lhs) + frame.get(stmt.rhs),)

    @interp.impl(stmts.Sub)
    def sub(self, interp, frame: interp.Frame, stmt: stmts.Sub):
        return (frame.get(stmt.lhs) - frame.get(stmt.rhs),)

    @interp.impl(stmts.Mult)
    def mult(self, interp, frame: interp.Frame, stmt: stmts.Mult):
        return (frame.get(stmt.lhs) * frame.get(stmt.rhs),)

    @interp.impl(stmts.Div)
    def div(self, interp, frame: interp.Frame, stmt: stmts.Div):
        return (frame.get(stmt.lhs) / frame.get(stmt.rhs),)

    @interp.impl(stmts.Mod)
    def mod(self, interp, frame: interp.Frame, stmt: stmts.Mod):
        return (frame.get(stmt.lhs) % frame.get(stmt.rhs),)

    @interp.impl(stmts.BitAnd)
    def bit_and(self, interp, frame: interp.Frame, stmt: stmts.BitAnd):
        return (frame.get(stmt.lhs) & frame.get(stmt.rhs),)

    @interp.impl(stmts.BitOr)
    def bit_or(self, interp, frame: interp.Frame, stmt: stmts.BitOr):
        return (frame.get(stmt.lhs) | frame.get(stmt.rhs),)

    @interp.impl(stmts.BitXor)
    def bit_xor(self, interp, frame: interp.Frame, stmt: stmts.BitXor):
        return (frame.get(stmt.lhs) ^ frame.get(stmt.rhs),)

    @interp.impl(stmts.LShift)
    def lshift(self, interp, frame: interp.Frame, stmt: stmts.LShift):
        return (frame.get(stmt.lhs) << frame.get(stmt.rhs),)

    @interp.impl(stmts.RShift)
    def rshift(self, interp, frame: interp.Frame, stmt: stmts.RShift):
        return (frame.get(stmt.lhs) >> frame.get(stmt.rhs),)

    @interp.impl(stmts.FloorDiv)
    def floor_div(self, interp, frame: interp.Frame, stmt: stmts.FloorDiv):
        return (frame.get(stmt.lhs) // frame.get(stmt.rhs),)

    @interp.impl(stmts.Pow)
    def pow(self, interp, frame: interp.Frame, stmt: stmts.Pow):
        return (frame.get(stmt.lhs) ** frame.get(stmt.rhs),)

    @interp.impl(stmts.MatMult)
    def mat_mult(self, interp, frame: interp.Frame, stmt: stmts.MatMult):
        return (frame.get(stmt.lhs) @ frame.get(stmt.rhs),)



from kirin import lowering
from kirin.prelude import python_no_opt
from kirin.dialects import cf, func

lower = lowering.Python(python_no_opt)


def test_pass():
    def nobody():
        pass

    code = lower.python_function(nobody)
    assert isinstance(code, func.Function)
    assert isinstance(code.body.blocks[-1].last_stmt, func.Return)

    def branch_pass():
        if True:
            pass
        else:
            pass

    code = lower.python_function(branch_pass)
    code.print()
    assert isinstance(code, func.Function)
    assert isinstance(code.body.blocks[0].last_stmt, func.Return)


def test_basic_ifelse():
    def single(n):
        if n == 0:
            return 1
        else:
            return n

    code = lower.python_function(single)
    assert isinstance(code, func.Function)
    assert len(code.body.blocks) == 3
    assert isinstance(code.body.blocks[0].last_stmt, cf.ConditionalBranch)
    assert code.body.blocks[0].last_stmt.then_successor is code.body.blocks[1]
    assert code.body.blocks[0].last_stmt.else_successor is code.body.blocks[2]
    assert isinstance(code.body.blocks[1].last_stmt, func.Return)
    assert isinstance(code.body.blocks[2].last_stmt, func.Return)

    def single_2(n):
        if n == 0:
            n + 1
        else:
            return n

    code = lower.python_function(single_2)
    code.print()
    assert isinstance(code, func.Function)
    assert len(code.body.blocks) == 3
    assert isinstance(code.body.blocks[0].last_stmt, cf.ConditionalBranch)
    assert code.body.blocks[0].last_stmt.then_successor is code.body.blocks[1]
    assert code.body.blocks[0].last_stmt.else_successor is code.body.blocks[2]
    assert isinstance(code.body.blocks[1].last_stmt, func.Return)
    assert isinstance(code.body.blocks[2].last_stmt, func.Return)

    def single_3(n):
        if n == 0:
            n = n + 1
        else:
            n = n + 2

    code = lower.python_function(single_3)
    assert isinstance(code, func.Function)
    assert len(code.body.blocks) == 4
    assert isinstance(code.body.blocks[0].last_stmt, cf.ConditionalBranch)
    assert code.body.blocks[0].last_stmt.then_successor is code.body.blocks[1]
    assert code.body.blocks[0].last_stmt.else_successor is code.body.blocks[2]
    assert isinstance(code.body.blocks[1].last_stmt, cf.Branch)
    assert code.body.blocks[1].last_stmt.successor is code.body.blocks[3]
    assert isinstance(code.body.blocks[2].last_stmt, cf.Branch)
    assert code.body.blocks[2].last_stmt.successor is code.body.blocks[3]
    assert isinstance(code.body.blocks[3].last_stmt, func.Return)

    def single_4(n):
        if n == 0:
            n = n + 1

    code = lower.python_function(single_4)
    assert isinstance(code, func.Function)
    assert len(code.body.blocks) == 3
    assert isinstance(code.body.blocks[0].last_stmt, cf.ConditionalBranch)
    assert code.body.blocks[0].last_stmt.then_successor is code.body.blocks[1]
    assert code.body.blocks[0].last_stmt.else_successor is code.body.blocks[2]
    assert isinstance(code.body.blocks[1].last_stmt, cf.Branch)
    assert code.body.blocks[1].last_stmt.successor is code.body.blocks[2]
    assert isinstance(code.body.blocks[2].last_stmt, func.Return)


def test_recursive_ifelse():
    def multi(n):
        if n == 0:
            return 1
        else:
            if n < 5:
                assert n < 6, "n must be less than 10"
            else:
                return n

    code = lower.python_function(multi)
    code.print()
    assert isinstance(code, func.Function)
    assert len(code.body.blocks) == 5
    assert isinstance(code.body.blocks[0].last_stmt, cf.ConditionalBranch)
    assert code.body.blocks[0].last_stmt.then_successor is code.body.blocks[1]
    assert code.body.blocks[0].last_stmt.else_successor is code.body.blocks[2]
    assert isinstance(code.body.blocks[1].last_stmt, func.Return)
    assert isinstance(code.body.blocks[2].last_stmt, cf.ConditionalBranch)
    assert code.body.blocks[2].last_stmt.then_successor is code.body.blocks[3]
    assert code.body.blocks[2].last_stmt.else_successor is code.body.blocks[4]
    assert isinstance(code.body.blocks[3].last_stmt, func.Return)
    assert isinstance(code.body.blocks[4].last_stmt, func.Return)



from dialect import dialect

from rewrite import RandomWalkBranch
from kirin.ir import dialect_group
from kirin.prelude import basic_no_opt
from kirin.rewrite import Walk, Fixpoint
from kirin.passes.fold import Fold


# create our own food dialect, it runs a random walk on the branches
@dialect_group(basic_no_opt.add(dialect))
def food(self):

    fold_pass = Fold(self)

    def run_pass(mt, *, fold=True):
        Fixpoint(Walk(RandomWalkBranch())).rewrite(mt.code)

        # add const fold
        if fold:
            fold_pass(mt)

    return run_pass



from kirin import types, interp
from kirin.dialects.py.binop import Add

from .stmts import New, Append
from ._dialect import dialect


@dialect.register
class ListMethods(interp.MethodTable):

    @interp.impl(New)
    def new(self, interp, frame: interp.Frame, stmt: New):
        return (list(frame.get_values(stmt.values)),)

    @interp.impl(Add, types.PyClass(list), types.PyClass(list))
    def add(self, interp, frame: interp.Frame, stmt: Add):
        return (frame.get(stmt.lhs) + frame.get(stmt.rhs),)

    @interp.impl(Append)
    def append(self, interp, frame: interp.Frame, stmt: Append):
        return (frame.get(stmt.list_).append(frame.get(stmt.value)),)



import pytest

from kirin.prelude import basic_no_opt


def test_basic():
    @basic_no_opt
    def single(n):
        if n == 0:
            return 1
        else:
            return n

    assert single(0) == 1
    assert single(1) == 1
    assert single(2) == 2

    @basic_no_opt
    def single_2(n):
        if n == 0:
            n + 1  # type: ignore
        else:
            return n

    assert single_2(0) is None
    assert single_2(1) == 1
    assert single_2(2) == 2

    @basic_no_opt
    def single_3(n):
        if n == 0:
            n = n + 1
        else:
            n = n + 2

    assert single_3(0) is None
    assert single_3(1) is None

    @basic_no_opt
    def single_4(n):
        if n == 0:
            n = n + 1

    assert single_4(0) is None
    assert single_4(1) is None


def test_assert():

    @basic_no_opt
    def multi(n):
        if n == 0:
            return 1
        else:
            if n < 5:
                assert n < 2, "n must be less than 2"
            else:
                return n

    assert multi(0) == 1
    assert multi(1) is None

    with pytest.raises(AssertionError):
        multi(2)


def test_recursive():
    @basic_no_opt
    def factorial(n):
        if n == 0:
            return 1
        else:
            return n * factorial(n - 1)

    assert factorial(0) == 1
    assert factorial(1) == 1
    assert factorial(2) == 2
    assert factorial(3) == 6


def test_kw_call():
    @basic_no_opt
    def callee(n, m):
        return n, m

    @basic_no_opt
    def caller(n, m):
        return callee(m=m, n=n)

    assert caller(1, 2) == (1, 2)



from kirin import types
from kirin.prelude import basic_no_opt
from kirin.dialects import py


@basic_no_opt
def explicit_slice():
    x = slice(1, 2, 3)
    y = slice(1, 2)
    z = slice(1)
    return x, y, z


@basic_no_opt
def wrong_slice():
    x = slice(None, None, None)
    y = slice(None, None, 1)
    return x, y


def test_explicit_slice():
    stmt: py.slice.Slice = explicit_slice.code.body.blocks[0].stmts.at(3)
    assert stmt.start.type.is_subseteq(types.Int)
    assert stmt.stop.type.is_subseteq(types.Int)
    assert stmt.step.type.is_subseteq(types.Int)
    assert stmt.result.type.is_subseteq(types.Slice[types.Int])

    stmt: py.slice.Slice = explicit_slice.code.body.blocks[0].stmts.at(7)
    assert stmt.start.type.is_subseteq(types.Int)
    assert stmt.stop.type.is_subseteq(types.Int)
    assert stmt.step.type.is_subseteq(types.NoneType)
    assert stmt.result.type.is_subseteq(types.Slice[types.Int])

    stmt: py.slice.Slice = explicit_slice.code.body.blocks[0].stmts.at(11)
    assert stmt.start.type.is_subseteq(types.NoneType)
    assert stmt.stop.type.is_subseteq(types.Int)
    assert stmt.step.type.is_subseteq(types.NoneType)
    assert stmt.result.type.is_subseteq(types.Slice[types.Int])


def test_wrong_slice():
    stmt: py.slice.Slice = wrong_slice.code.body.blocks[0].stmts.at(3)
    assert stmt.result.type.is_subseteq(types.Bottom)

    stmt: py.slice.Slice = wrong_slice.code.body.blocks[0].stmts.at(7)
    assert stmt.result.type.is_subseteq(types.Bottom)



import kirin.prelude


def iter_non_pure():
    @kirin.prelude.basic
    def loop(a: str):
        out = []
        for i in range(4):
            # for i in [1,1,2,3]:  # Same result with this line instead
            out = out + [a]
        return out

    x = loop("a")
    assert x == ["a", "a", "a", "a"]

    x = loop("b")
    assert x == ["b", "b", "b", "b"]

    x = loop("c")
    assert x == ["c", "c", "c", "c"]



"""Python dialects module.

This module contains a set of dialects that represent
different fractions of the Python language. The dialects
are designed to be used in a union to represent the
entire Python language.
"""

from . import (
    cmp as cmp,
    len as len,
    attr as attr,
    base as base,
    list as list,
    binop as binop,
    range as range,
    slice as slice,
    tuple as tuple,
    unary as unary,
    assign as assign,
    boolop as boolop,
    unpack as unpack,
    builtin as builtin,
    constant as constant,
    indexing as indexing,
    iterable as iterable,
)
from .len import Len as Len
from .attr import GetAttr as GetAttr
from .range import Range as Range
from .slice import Slice as Slice
from .assign import Alias as Alias, SetItem as SetItem
from .boolop import Or as Or, And as And
from .builtin import Abs as Abs, Sum as Sum
from .constant import Constant as Constant
from .indexing import GetItem as GetItem, PyGetItemLike as PyGetItemLike
from .cmp.stmts import *  # noqa: F403
from .list.stmts import Append as Append
from .binop.stmts import *  # noqa: F403
from .unary.stmts import *  # noqa: F403



import pytest

from kirin.prelude import basic


class Foo:

    def some_kernel(self):
        @basic(verify=False)
        def goo(x: int):
            return x

        return goo

    def another_kernel(self):
        @basic(verify=False)
        def goo(x: int):
            kernel = self.some_kernel()
            kernel(x)

        return goo


def test_call_method_error():
    foo = Foo()
    goo = foo.another_kernel()

    with pytest.raises(Exception):
        goo.verify_type()



from kirin import ir, types, lowering
from kirin.decl import info, statement

from ._dialect import dialect

T = types.TypeVar("T")


@statement
class UnaryOp(ir.Statement):
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    value: ir.SSAValue = info.argument(T, print=False)
    result: ir.ResultValue = info.result(T)


@statement(dialect=dialect)
class UAdd(UnaryOp):
    name = "uadd"


@statement(dialect=dialect)
class USub(UnaryOp):
    name = "usub"


@statement(dialect=dialect)
class Not(UnaryOp):
    name = "not"


@statement(dialect=dialect)
class Invert(UnaryOp):
    name = "invert"
